"""PSUtils document reader classes.

Copyright (c) Reuben Thomas 2023-2025.
Released under the GPL version 3, or (at your option) any later version.
"""

import re
from typing import IO

from pypdf import PdfReader as PdfReaderBase
from pypdf._utils import StrByteType

from .types import Rectangle
from .warnings import die


class PdfReader(PdfReaderBase):
    def __init__(
        self,
        stream: StrByteType,
        strict: bool = False,
        password: str | bytes | None = None,
    ) -> None:
        super().__init__(stream, strict, password)
        assert len(self.pages) > 0
        mediabox = self.pages[0].mediabox
        self.size = Rectangle(mediabox.width, mediabox.height)
        self.size_guessed = False


size_keywords = (
    b"DocumentMedia",
    b"PageBoundingBox",
    b"HiResBoundingBox",
    b"BoundingBox",
)


# FIXME: Store lists of lines, not file offsets.
class PsReader:
    def __init__(self, infile: IO[bytes]) -> None:
        self.infile = infile
        self.headerpos: int = 0
        self.pagescmt: int = 0
        self.endsetup: int = 0
        self.procset_pos: range = range(0, 0)  # pstops procset location
        self.num_pages: int = 0
        self.sizeheaders: list[int] = []
        self.pageptr: list[int] = []
        self.size = None
        self.size_guessed = False

        nesting = 0
        self.infile.seek(0)
        record, next_record, buffer = 0, 0, None
        file_sizes = {}
        for buffer in self.infile:
            next_record += len(buffer)
            if buffer.startswith(b"%%"):
                keyword, value = self.comment(buffer)
                if keyword is not None:
                    # If input paper size is not set, try to read it
                    if (
                        self.headerpos == 0
                        and self.size is None
                        and keyword in size_keywords
                    ):
                        assert value is not None
                        words = value.split(b" ")
                        if keyword == b"DocumentMedia" and len(words) > 2:
                            w = words[1].decode("utf-8", "ignore")
                            h = words[2].decode("utf-8", "ignore")
                            try:
                                file_sizes[keyword] = Rectangle(float(w), float(h))
                            except ValueError:
                                pass
                        elif len(words) == 4:
                            llx = words[0].decode("utf-8", "ignore")
                            lly = words[1].decode("utf-8", "ignore")
                            urx = words[2].decode("utf-8", "ignore")
                            ury = words[3].decode("utf-8", "ignore")
                            try:
                                file_sizes[keyword] = Rectangle(
                                    float(urx) - float(llx), float(ury) - float(lly)
                                )
                            except ValueError:
                                pass
                    if nesting == 0 and keyword == b"Page":
                        self.pageptr.append(record)
                    elif self.headerpos == 0 and (
                        keyword in size_keywords or keyword == b"DocumentPaperSizes"
                    ):
                        self.sizeheaders.append(record)
                    elif self.headerpos == 0 and keyword == b"Pages":
                        self.pagescmt = record
                    elif self.headerpos == 0 and keyword == b"EndComments":
                        self.headerpos = next_record
                    elif keyword in [
                        b"BeginDocument",
                        b"BeginBinary",
                        b"Begself.infile",
                    ]:
                        nesting += 1
                    elif keyword in [b"EndDocument", b"EndBinary", b"EndFile"]:
                        nesting -= 1
                    elif nesting == 0 and keyword == b"EndSetup":
                        self.endsetup = record
                    elif nesting == 0 and keyword == b"BeginProlog":
                        self.headerpos = next_record
                    elif nesting == 0 and buffer == b"%%BeginProcSet: PStoPS":
                        self.procset_pos = range(record, 0)
                    elif (
                        self.procset_pos.start > 0
                        and self.procset_pos.stop == 0
                        and keyword == b"EndProcSet"
                    ):
                        self.procset_pos = range(self.procset_pos.start, next_record)
                    elif nesting == 0 and keyword in [b"Trailer", b"EOF"]:
                        break
            elif self.headerpos == 0:
                self.headerpos = record
            record = next_record

        # If paper size was not already set, and we found a possible size in
        # the file, use it.
        if self.size is None and len(file_sizes) > 0:
            for keyword in size_keywords:
                file_size = file_sizes.get(keyword)
                if (
                    file_size is not None
                    and file_size.width != 0
                    and file_size.height != 0
                ):
                    self.size = file_size
                    if keyword in (b"BoundingBox", b"HiResBoundingBox"):
                        self.size_guessed = True
                    break
        self.num_pages = len(self.pageptr)
        self.pageptr.append(record)
        if self.endsetup == 0 or self.endsetup > self.pageptr[0]:
            self.endsetup = self.pageptr[0]

    # Return comment keyword and value if `line' is a DSC comment
    def comment(self, line: bytes) -> tuple[bytes, bytes] | tuple[None, None]:
        m = re.match(b"%%([^:]+):?\\s+?(.*\\S?)\\s*$", line)
        return (m[1], m[2]) if m else (None, None)


def document_reader(file: IO[bytes], file_type: str) -> PdfReader | PsReader:
    constructor: type[PdfReader] | type[PsReader]
    if file_type in (".ps", ".eps"):
        constructor = PsReader
    elif file_type == ".pdf":
        constructor = PdfReader
    else:
        die(f"incompatible file type `{file_type}'")
    return constructor(file)



"""psbook command.

Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import argparse
import sys
import warnings

from psutils.argparse import (
    HelpFormatter,
    PaperContext,
    add_basic_arguments,
    parserange,
    parsespecs,
)
from psutils.transformers import file_transform
from psutils.warnings import die, simple_warning


def get_parser() -> argparse.ArgumentParser:
    # Command-line arguments
    parser = argparse.ArgumentParser(
        description="Rearrange pages in a PDF or PostScript document into signatures.",
        formatter_class=HelpFormatter,
        usage="%(prog)s [OPTION...] [INFILE [OUTFILE]]",
        add_help=False,
        epilog="""
PAGES is a comma-separated list of pages and page ranges; see pstops(1)
for more details.
""",
    )
    warnings.showwarning = simple_warning(parser.prog)

    # Command-line parser
    parser.add_argument(
        "-s",
        "--signature",
        type=int,
        default=0,
        help="""\
number of pages per signature;
0 = all pages in one signature [default];
1 = do not rearrange the pages;
otherwise, a multiple of 4""",
    )
    add_basic_arguments(parser)

    return parser


def psbook(argv: list[str] = sys.argv[1:]) -> None:
    args = get_parser().parse_intermixed_args(argv)

    if args.signature > 1 and args.signature % 4 != 0:
        die("signature must be a multiple of 4")

    # Get number of pages
    paper_context = PaperContext()
    specs, modulo, flipping = parsespecs("0", paper_context)
    with file_transform(
        args.infile, args.outfile, None, None, specs, 0, False
    ) as transform:
        input_pages = transform.pages()

        def page_index_to_real_page(signature: int, page_number: int) -> int:
            real_page = page_number - page_number % signature
            page_on_sheet = page_number % 4
            recto_verso = (page_number % signature) // 2
            if page_on_sheet in (0, 3):
                real_page += signature - 1 - recto_verso
            else:
                real_page += recto_verso
            return real_page + 1

        # Adjust for signature size
        signature = args.signature
        if signature == 0:
            maxpage = input_pages + (4 - input_pages % 4) % 4
            signature = maxpage
        else:
            maxpage = input_pages + (signature - input_pages % signature) % signature

        # Compute page list
        page_list = []
        for page in range(maxpage):
            real_page = page_index_to_real_page(signature, page)
            page_list.append(str(real_page) if real_page <= input_pages else "_")

        # Rearrange pages
        transform.transform_pages(
            parserange(",".join(page_list)),
            flipping,
            False,
            False,
            False,
            modulo,
            args.verbose,
        )


if __name__ == "__main__":
    psbook()
